==> ampform-0.8.1.tar.gz: ampform-0.8.1/src/ampform/dynamics/builder.py <==
"""Build :mod:`.lineshape` with correct variable names and values."""

import inspect
from collections import OrderedDict
from typing import Callable, Dict, Optional, Tuple

import attr
import sympy as sp
from qrules.particle import Particle

from .lineshape import (
    BlattWeisskopf,
    breakup_momentum,
    relativistic_breit_wigner,
    relativistic_breit_wigner_with_ff,
)

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore


@attr.s(frozen=True)
class TwoBodyKinematicVariableSet:
    in_edge_inv_mass: sp.Symbol = attr.ib()
    out_edge_inv_mass1: sp.Symbol = attr.ib()
    out_edge_inv_mass2: sp.Symbol = attr.ib()
    helicity_theta: sp.Symbol = attr.ib()
    helicity_phi: sp.Symbol = attr.ib()
    angular_momentum: Optional[int] = attr.ib(default=None)


def create_non_dynamic(
    resonance: Particle, variable_pool: TwoBodyKinematicVariableSet
) -> Tuple[sp.Expr, Dict[sp.Symbol, float]]:
    # pylint: disable=unused-argument
    return (1, {})


def create_non_dynamic_with_ff(
    resonance: Particle, variable_pool: TwoBodyKinematicVariableSet
) -> Tuple[sp.Expr, Dict[sp.Symbol, float]]:
    angular_momentum = variable_pool.angular_momentum
    if angular_momentum is None:
        raise ValueError(
            "Angular momentum is not defined but is required in the form factor!"
        )
    q = breakup_momentum(
        variable_pool.in_edge_inv_mass,
        variable_pool.out_edge_inv_mass1,
        variable_pool.out_edge_inv_mass2,
    )
    meson_radius = sp.Symbol(f"d_{resonance.name}")
    return (
        BlattWeisskopf(q, meson_radius, angular_momentum),
        {meson_radius: 1},
    )


def create_relativistic_breit_wigner(
    resonance: Particle, variable_pool: TwoBodyKinematicVariableSet
) -> Tuple[sp.Expr, Dict[sp.Symbol, float]]:
    inv_mass = variable_pool.in_edge_inv_mass
    res_mass = sp.Symbol(f"m_{resonance.name}")
    res_width = sp.Symbol(f"Gamma_{resonance.name}")
    expression = relativistic_breit_wigner(inv_mass, res_mass, res_width)
    parameter_defaults = {
        res_mass: resonance.mass,
        res_width: resonance.width,
    }
    return expression, parameter_defaults


def create_relativistic_breit_wigner_with_ff(
    resonance: Particle, variable_pool: TwoBodyKinematicVariableSet
) -> Tuple[sp.Expr, Dict[sp.Symbol, float]]:
    if variable_pool.angular_momentum is None:
        raise ValueError(
            "Angular momentum is not defined but is required in the form factor!"
        )

    inv_mass = variable_pool.in_edge_inv_mass
    res_mass = sp.Symbol(f"m_{resonance.name}")
    res_width = sp.Symbol(f"Gamma_{resonance.name}")
    product1_inv_mass = variable_pool.out_edge_inv_mass1
    product2_inv_mass = variable_pool.out_edge_inv_mass2
    angular_momentum = variable_pool.angular_momentum
    meson_radius = sp.Symbol(f"d_{resonance.name}")

    expression = relativistic_breit_wigner_with_ff(
        inv_mass,
        res_mass,
        res_width,
        product1_inv_mass,
        product2_inv_mass,
        angular_momentum,
        meson_radius,
    )
    parameter_defaults = {
        res_mass: resonance.mass,
        res_width: resonance.width,
        meson_radius: 1,
    }
    return expression, parameter_defaults


class ResonanceDynamicsBuilder(Protocol):
    """Protocol that is used by `.set_dynamics`.

    Follow this `~typing.Protocol` when defining a builder function that is to
    be used by `.set_dynamics`. For an example, see the source code
    `.create_relativistic_breit_wigner`, which creates a
    `.relativistic_breit_wigner`.

    .. seealso:: :doc:`/usage/dynamics/custom`
    """

    def __call__(
        self, resonance: Particle, variable_pool: TwoBodyKinematicVariableSet
    ) -> Tuple[sp.Expr, Dict[sp.Symbol, float]]:
        ...


def verify_signature(builder: Callable) -> None:
    """Check signature of a builder function dynamically.

    Dynamically check whether a builder has the same signature as
    `.ResonanceDynamicsBuilder`. This function is needed because
    `typing.runtime_checkable` does only checks members and methods, not the
    signature of those methods.
    """
    signature = inspect.signature(builder)
    if signature.return_annotation != __EXPECTED.return_annotation:
        raise ValueError(
            f'Builder "{builder.__name__}" has return type {__EXPECTED.return_annotation};'
            f" expected {signature.return_annotation}"
        )
    expected_parameters = OrderedDict(__EXPECTED.parameters.items())
    del expected_parameters["self"]
    assert signature.return_annotation == __EXPECTED.return_annotation
    if signature.parameters != expected_parameters:
        raise ValueError(
            f'Builder "{builder.__name__}" has parameters\n'
            f"  {list(signature.parameters.values())}\n"
            "This should be\n"
            f"  {list(expected_parameters.values())}"
        )


__EXPECTED = inspect.signature(ResonanceDynamicsBuilder.__call__)

==> ampform-0.9.0.tar.gz: ampform-0.9.0/src/ampform/dynamics/builder.py <==
"""Build `~ampform.dynamics` with correct variable names and values."""

from typing import Dict, Optional, Tuple

import attr
import sympy as sp
from qrules.particle import Particle

from . import (
    BlattWeisskopfSquared,
    breakup_momentum_squared,
    relativistic_breit_wigner,
    relativistic_breit_wigner_with_ff,
)

# pyright: reportUnusedImport=false
from .decorator import verify_signature  # noqa: F401

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore


@attr.s(frozen=True)
class TwoBodyKinematicVariableSet:
    """Data container for the essential variables of a two-body decay.

    This data container is inserted into a `.ResonanceDynamicsBuilder`, so that
    it can build some lineshape expression from the `.dynamics` module. It also
    allows to insert :doc:`custom dynamics </usage/dynamics/custom>` into the
    amplitude model.
    """

    in_edge_inv_mass: sp.Symbol = attr.ib()
    out_edge_inv_mass1: sp.Symbol = attr.ib()
    out_edge_inv_mass2: sp.Symbol = attr.ib()
    helicity_theta: sp.Symbol = attr.ib()
    helicity_phi: sp.Symbol = attr.ib()
    angular_momentum: Optional[int] = attr.ib(default=None)


def create_non_dynamic(
    resonance: Particle, variable_pool: TwoBodyKinematicVariableSet
) -> Tuple[sp.Expr, Dict[sp.Symbol, float]]:
    # pylint: disable=unused-argument
    return (1, {})


def create_non_dynamic_with_ff(
    resonance: Particle, variable_pool: TwoBodyKinematicVariableSet
) -> Tuple[sp.Expr, Dict[sp.Symbol, float]]:
    """Generate (only) a Blatt-Weisskopf form factor for a two-body decay.

    Returns the `~sympy.functions.elementary.miscellaneous.sqrt` of a
    `.BlattWeisskopfSquared`.
    """
    angular_momentum = variable_pool.angular_momentum
    if angular_momentum is None:
        raise ValueError(
            "Angular momentum is not defined but is required in the form factor!"
        )
    q_squared = breakup_momentum_squared(
        s=variable_pool.in_edge_inv_mass ** 2,
        m_a=variable_pool.out_edge_inv_mass1,
        m_b=variable_pool.out_edge_inv_mass2,
    )
    meson_radius = sp.Symbol(f"d_{resonance.name}")
    form_factor_squared = BlattWeisskopfSquared(
        angular_momentum,
        z=q_squared * meson_radius ** 2,
    )
    return (
        sp.sqrt(form_factor_squared),
        {meson_radius: 1},
    )


def create_relativistic_breit_wigner(
    resonance: Particle, variable_pool: TwoBodyKinematicVariableSet
) -> Tuple[sp.Expr, Dict[sp.Symbol, float]]:
    """Create a `.relativistic_breit_wigner` for a two-body decay."""
    inv_mass = variable_pool.in_edge_inv_mass
    res_mass = sp.Symbol(f"m_{resonance.name}")
    res_width = sp.Symbol(f"Gamma_{resonance.name}")
    expression = relativistic_breit_wigner(
        s=inv_mass ** 2,
        mass0=res_mass,
        gamma0=res_width,
    )
    parameter_defaults = {
        res_mass: resonance.mass,
        res_width: resonance.width,
    }
    return expression, parameter_defaults


def create_relativistic_breit_wigner_with_ff(
    resonance: Particle, variable_pool: TwoBodyKinematicVariableSet
) -> Tuple[sp.Expr, Dict[sp.Symbol, float]]:
    """Create a `.relativistic_breit_wigner_with_ff` for a two-body decay."""
    if variable_pool.angular_momentum is None:
        raise ValueError(
            "Angular momentum is not defined but is required in the form factor!"
        )

    inv_mass = variable_pool.in_edge_inv_mass
    res_mass = sp.Symbol(f"m_{resonance.name}")
    res_width = sp.Symbol(f"Gamma_{resonance.name}")
    product1_inv_mass = variable_pool.out_edge_inv_mass1
    product2_inv_mass = variable_pool.out_edge_inv_mass2
    angular_momentum = variable_pool.angular_momentum
    meson_radius = sp.Symbol(f"d_{resonance.name}")

    expression = relativistic_breit_wigner_with_ff(
        s=inv_mass ** 2,
        mass0=res_mass,
        gamma0=res_width,
        m_a=product1_inv_mass,
        m_b=product2_inv_mass,
        angular_momentum=angular_momentum,
        meson_radius=meson_radius,
    )
    parameter_defaults = {
        res_mass: resonance.mass,
        res_width: resonance.width,
        meson_radius: 1,
    }
    return expression, parameter_defaults


class ResonanceDynamicsBuilder(Protocol):
    """Protocol that is used by `.set_dynamics`.

    Follow this `~typing.Protocol` when defining a builder function that is to
    be used by `.set_dynamics`. For an example, see the source code
    `.create_relativistic_breit_wigner`, which creates a
    `.relativistic_breit_wigner`.

    .. seealso:: :doc:`/usage/dynamics/custom`
    """

    def __call__(
        self, resonance: Particle, variable_pool: TwoBodyKinematicVariableSet
    ) -> Tuple[sp.Expr, Dict[sp.Symbol, float]]:
        ...
